#!/usr/bin/env python
# -*- coding: iso-8859-1 -*-
"""
Dieses Skript implementiert so etwas wie einen kleinen Proxy-Server
ohne Proxy.  Er bersetzt Seiten mit einen externen Filter und
biegt zustzlich URLs so um, dass sie -- bei Bildern -- auf den
Originalserver verweisen oder -- bei Links -- wieder zurck auf
dieses Skript.
"""

import re
import timeoutsocket
import common
import urllib
import urllib2
import urlparse
import operator
import os

# Die Filter sind externe Programme.  Als Schlssel dienen hier die
# Schlsselwrter, die das Skript aus externen Quellen nimmt
langFilters = {
	'schwob_url': 'webschwob',
	'sax_url': 'websax',
	'frank_url': 'frankifier',
}

# Dieses Dictionary wird von checkPermission verwendet.  Ist
# der Host der Zielurl oder ein Teil davon in diesem Dicitionary
# aufgefhrt, wird die bersetzung verweigert
spinner = {
	"www.test.bla": 1,
}

noTranslateExtensions = {
".gif": 1,
".jpg": 1,
".pdf": 1,
".ps": 1,
".pdb": 1,
".jpeg": 1,
".mpg": 1,
}


# Ich brauche dringend einen timeout fr die http-Verbindungen,
# und signal.alarm bringt den apache durcheinander, jedenfalls,
# wenn das hier unter mod_python lebt.  Unter moderneren 
# Python-Versionen gibts sowas eingebaut, aber darauf mchte ich
# mich nicht verlassen.
timeoutsocket.setDefaultSocketTimeout(30)


def checkPermission(targURL):
	"""Prft, ob fr targURL eine bersetzung vorgenommen werden soll/darf.
	Wir prfen im Augenblick nur den Host Part.  Vgl. auch das spinner-dict
	oben.
	"""
	try:
		hostname = re.sub(":.*", "", urlparse.urlparse(targURL)[1])
		canonHost = hostname.lower()
		if spinner.has_key(canonHost) or reduce(operator.__or__, 
				map(lambda a, d=spinner: d.has_key(a), canonHost.split("."))):
			return 1
	except IndexError:
		pass


def workAroundUrllibBug(url):
	"""Umgeht einen Fehler in der urllib, der eine korrekte Interpretation
	von URLs wie http://host?bla verhindert.  Es gibt einen Bug Report
	dazu, aber offenbar hat niemand Lust, das an der Quelle in Ordnung
	zu bringen.
	"""
	if not '?' in url:
		return url
	pre, post = url.split("?", 1)
	if "/" in pre[7:]:
		return url
	return pre+"/?"+post


import urllib2

class SchwobErrHandler(urllib2.BaseHandler):
	"""Diese Klasse liefert etwas przisere Fehler als der Orinialhandler
	der urllib2.
	"""
	def http_error_401(self, req, fp, code, msg, hdrs):
		raise common.Error("Die Seite %s ist passwortgeschtzt.  Der"
			" USP bersetzt so etwas nicht.")%req.get_full_url()

	def http_error_403(self, req, fp, code, msg, hdrs):
		raise common.Error("Die Seite %s ist fr mich verboten.  Knnte sein,"
			" dass die Seitenbetreiber kein Schwbisch mgen.")%(
			req.get_full_url())
	
	def http_error_404(self, req, fp, code, msg, hdrs):
		raise common.Error("Fr die Seite %s habe ich ein 404 bekommen."
			" Das bedeutet meistens, dass es die Seite nicht mehr gibt,"
			" es kann aber auch sein, dass ich was falsch verstanden habe."
			" So oder so, die Seite gibt es so nicht.")%(
			req.get_full_url())


class SchwobOpenerDirector(urllib2.OpenerDirector):
	"""Hauptziel hier ist ein korrekter User-agent-String.  Darber
	hinaus unterdrcke ich hier auch Handler fr file, ftp und
	Freunde, mit denen irgendwer Unfug machen knnte.
	"""
	def __init__(self):
		urllib2.OpenerDirector.__init__(self)
		self.addheaders = [('User-agent', "Schwobifying Proxy, %s"%
			common.proxyHomeURL)]
		for handler in [urllib2.HTTPHandler(), SchwobErrHandler(),
			urllib2.HTTPDefaultErrorHandler(), urllib2.HTTPRedirectHandler(), 
			urllib2.UnknownHandler()]:
			self.add_handler(handler)

urllib2.install_opener(SchwobOpenerDirector())


def urljoin(url1, url2):
	"""Noch ein Versuch, url1 und url2 sinnvoll zu verbinden -- was urljoin
	tut, hat mir nicht immer gepasst.  Aber zugegeben, das hier ist
	noch grerer Unfug.  Dafr funktioniert es meistens, auch in der
	Gegenwart von index.htmls.
	"""
	u1 = urlparse.urlparse(url1)
	u2 = urlparse.urlparse(url2)
	if u2[0]:
		# url2 hat scheme und ist also offentichtlich komplett
		return url2
	if u2[1]:
		# urlparse deklariert alles ohne einen / davor als host, was
		# wenigstens hier Quatsch ist
		u2 = ('','',os.path.join(u2[1],u2[2]),u2[3],u2[4],u2[5])
	
	# die zweite URL knnte ein Pfad sein -- oder nur ein Tag
	if not u2[2] and u2[5]:
		return urlparse.urlunparse((u1[0],u1[1],u1[2],u1[3],u1[4],u2[5]))
	if not u2[2]:
		u2 = (u2[0],u2[1],"/",u2[3],u2[4],u2[5])
	newpath = os.path.join(os.path.dirname(u1[2]),u2[2])
	return urlparse.urlunparse((u1[0],u1[1],re.sub("^/","",newpath),
		u2[3],u2[4],u2[5]))


def raiseTimeout(signum, stackframe):
	raise common.TimeoutError, ("Die Maschine am anderen Ende hat sich"
		" nicht rechtzeitig gemeldet")


class Document:
	"""Diese Klasse verkapselt ein Dokument samt seines bersetzungsprozesses.
	Sie wurde etwas roh aus ihrer Heimat -- dem UNiMUT -- gerissen und ist
	deshalb vielleicht etwas weniger klar als man wnschen knnte...

	Darber hinaus operieren wir vorne und hinten mit regulren Ausdrcken
	auf HTML.  Es wre weit besser, etwas wie BeautifulSoup zu verwenden
	-- aber das berlassen wir anderen Leuten.
	"""
	def __init__(self, request):
		"""request ist ein Request-Objekt (vgl. unten).
		"""
		self.request = request
		self.proxyURL = common.proxyHomeURL
		for processor in langFilters.keys():
			if request.form.has_key(processor):
				self.processor = processor
				break
		else:
			raise common.Error("Interner Fehler.  Yikes.")

		self.showFooter = 1
		if request.form.has_key("nofooter"):
			self.showFooter = 0

		if type(request.form[self.processor])==type([]):
			raise common.Error("""Bitte den schwob_url-Parameter
				nur einmal angeben.  Der wahrscheinlichste Grund dieses
				Problems ist die Verwendung des bersetzten schwob-URL-Formulars.
				Bitte das
				<a href="%s">unbersetzte Formular</a>
				verwenden."""%self.proxyURL)
		self.sourceURL = request.form[self.processor].value
		if self.sourceURL.find("//")==-1:
			self.sourceURL = "http://"+self.sourceURL

		self._getDoc()

		self._doMeta()
		mat = re.search(r'''(?i)<base href=['"]?([^\s'">]+)''', self.data)
		if mat:
			self.sourceURL = mat.group(1)
			self.data = re.sub("(?is)<base href.*?>", "", self.data)
		if not self.isData:
			self._fixurls()
			self._fixBody()
			self._filterThroughWebschwob()

	def _filterThroughWebschwob(self):
		"""besorgt das eigentliche Filtern durch den externen Prozess.
		Wir machen uns das Leben hier leicht und filtern in eine
		temporre Datei, die wir dann am Stck lesen knnen.  Das
		erspart uns eine bidirektionale Pipe, die immer unerfreulich
		ist.

		Etwas trickreich ist hier, dass wir komplette Teile nicht
		filtern wollen.  Dazu gehren script- und style-Bereiche.  Es
		wre denkbar, das den externen Programmen zu berlassen, aber
		hier tun wir das selbst.  Dabei wird einfach jeweils etwas wie
		eine entity in den zu bersetzenden Text geschrieben und diese
		Entities nachher wieder ersetzt.
		"""
		import tempfile
		fn = tempfile.mktemp()
		noProcessChunks = {}
		processChunks = []
		noProcessOpeners = re.compile(
			"""<(script|style|nomangle)""", re.I)
		noProcessClosers = re.compile("""</(script|style|nomangle)>""", re.I)

		pos = 0
		while 1:
			matStart = noProcessOpeners.search(self.data, pos)
			if not matStart:
				processChunks.append(self.data[pos:])
				break
			matEnd = noProcessClosers.search(self.data, matStart.end())
			if not matEnd:
				processChunks.append(self.data[pos:])
				break
			processChunks.append(self.data[pos:matStart.start()])
			label = "&schwobachunk%d;"%pos
			processChunks.append(label)
			noProcessChunks[label] = self.data[matStart.start():matEnd.end()]
			pos = matEnd.end()
		try:
			open(fn, "w").write("\n".join(processChunks))
			self.data = os.popen(os.path.join(common.physPath,
				langFilters[self.processor])+" < %s"%fn).read()
			for label in noProcessChunks.keys():
				self.data = self.data.replace(label, noProcessChunks[label])
		finally:
			try:
				os.unlink(fn)
			except os.error:
				pass

	def _doMeta(self):
		"""interpretiert die Metas, die NutzerInnen in Seiten einbauen
		knnen, und setzt ggf. Attribute dementsprechend.
		"""
		mat = re.search("""(?is)<meta name=["']schwobcontrol["'] ([^>]*?)>""",
			self.data)
		if not mat:
			return
		if mat.group(1).find("nofooter")!=-1:
			self.showFooter = 0
		if mat.group(1).find("notranslate")!=-1:
			self._raiseForbidError()

	def _mangleURL(self, matchobj):
		"""wird als callback von re.sub aus aufgerufen.  matchobj braucht
		mindestens drei Gruppen: den Teiltag vor der URL, die URL selbst und
		den Teiltag nach der URL.  Zurckgegeben wird der komplette Tag mit
		der umgebogenen URL; mangleURL sorgt dafr, dass auch das Ziel des 
		Links durch den Schobifier gejagt wird.
		"""
		if matchobj.group(2).find("schwobnomangle")!=-1:
			return matchobj.group(0)
		if matchobj.group(2).startswith("mailto:"):
			return matchobj.group(0)
		url = matchobj.group(2)
		if noTranslateExtensions.has_key(os.path.splitext(url)[1]):
			return matchobj.group(0)
		else:
			newTag = matchobj.group(1)+self.proxyURL+("/schwob?%s="%self.processor
				)+urllib.quote_plus(urljoin(
				self.sourceURL, url)).replace("%23", "#")+matchobj.group(3)
			return newTag

	def _mangleActionURL(self, matchobj):
		"""biegt eine Action-URL aus einem Form um.  Idee ist
		hier, einerseits die Zielurl des Forms auf uns weisen
		zu lassen, andererseits die wirkliche Zielurl in einem
		zustzlichen (versteckten) Parameter des Formulars
		unterzubringen.  Mailto geht natrlich aus technischen
		Grnden nicht, POST wollen wir nicht, weil (a) die
		gegenwrtige Implementation mit sowas nicht umgehen
		kann und (b) wir nicht wirklich ein offener Proxy fr
		jeden denkbaren Unfug sein wollen.
		"""
		if matchobj.group(2).startswith("mailto:"):
			return matchobj.group(0)
		elif re.search("(?i)POST", matchobj.group(0)):
			return matchobj.group(0)
		else:
			return matchobj.group(1)+self.proxyURL+"/schwob"+matchobj.group(3)+\
				('<input type="hidden" name="%s"'%self.processor
				)+'value="%s">'%urljoin(self.sourceURL, matchobj.group(2))

	def _absolutifyURL(self, matchobj):
		"""biegt eine URL so um, dass sie absolut auf den Ursprungsserver
		zeigt (fr Bilder, Javascript und hnliches).
		"""
		return matchobj.group(1)+urljoin(self.sourceURL,
			matchobj.group(2))+matchobj.group(3)

	def _hackJavascript(self, matchobj):
		"""Ein verzweifelter Versuch, mit Javascript-Krampf klarzukommen --
		soll mit einem Matchobj auf eine (<script...>)(bla)(</script>)
		RE gerufen werden und versucht, URLs innerhalb von Javascript
		umzubiegen.  Wirklich gut funktioniert das natrlich nicht.
		"""
		return re.sub(r"""(["'])([^"']+?\.(?:gif|jpe?g))(["'])""",
				self._absolutifyURL,
			re.sub(r"""(?si)(href\s*=\s*["'])([^"'(]+)(["'])""",
				self._mangleURL, 
			re.sub(r"""(?si)(src\s*=\s*["'])([^"']+)(["'])""",
				self._absolutifyURL, matchobj.group())))

	def _fixurls(self):
		"""biegt die URLs in data auf den Proxy um bzw. wandelt sie in
		absolute um, weil die Dokumentenbasis anders wird.
		"""
		self.data = \
			re.sub("(?i)</head>",
			"<base href='%s'></head>"%self.sourceURL,
			re.sub("\x254", "</script>",
			re.sub(r"(?si)(<script[^>]+?javascript[^>]*>)([^\x254])(\x254)",
			self._hackJavascript,
			re.sub("</script>", "\x254",
			re.sub(
			r"""(background-image:\s*url\()([^)]+)(\))""",
			self._absolutifyURL,
			re.sub(
			r"""(?si)(onmouse(?:over|out|up|down)\s*=\s*["'][^>]*?\.src="""
			r"""["'])([^"']+)(["'])""",
			self._absolutifyURL,
			re.sub(
			r"""(?si)(<link\s[^>]*?href\s*=\s*["']?)([^"' >]+?)"""
			"""(["' >][^>]*>)""",
			self._absolutifyURL,
			re.sub(
			r"""(?si)(<(?:body|td|table)\s[^>]*(?:backg?round|td)\s*=\s*["']?)"""
			r"""([^"' >]+?)(["' >][^>]*>)""",
			self._absolutifyURL,
			re.sub(
			r"""(?si)(<applet\s[^>]*?code\s*=\s*["']?)([^"' >]+?)"""
			r"""(["' ].*?>|>)""",
			self._absolutifyURL, 
			re.sub("""(?si)(<meta[^>]+refresh[^>]+url\s*=\s*["']?)([^"' >]+?)"""
			r"""(["' ].*?>|>)""",
			self._mangleURL,
			re.sub(
			r"""(?si)(<(?:input|img|script)\s[^>]*?src\s*=\s*["']?)([^"' >]+?)"""
			r"""(["' ].*?>|>)""",
			self._absolutifyURL,
			re.sub(r"""(?si)(<frame\s[^>]*src\s*=\s*["']?)([^"' >]+?)"""
			"""(["' >][^>]*>)""",
			self._mangleURL,
			re.sub(r"""(?si)(<form\s[^>]*action\s*=\s*["']?)([^"' >]+?)"""
			"""(["' >][^>]*>)""",
			self._mangleActionURL,
			re.sub(r"""(?si)(<(?:a|area)\s[^>]*href\s*=\s*["']?)([^"' (>]+?)"""
			"""(["' >][^>]*>)""",
			self._mangleURL, self.data))))))))))))))

	def _fixBody(self):
		"""lscht schlieende HTML- oder BODY-Tags -- wir wollen da
		spter unseren eigenen Fu dranbasteln (getFooter).
		"""
		self.data = re.sub(r"(?i)\s*</HTML>\s*$", "",
			re.sub(r"(?i)\s*</BODY>\s*$", "", self.data))

	def getFooter(self):
		"""gibt Material fr das Ende der Seite zurck.
		"""
		if self.showFooter:
			return """<hr><p>Prozessiert von <a href="%(proxyURL)s">%(proxyURL)s</a>.
				<br>
				<font color="red">SeitenbetreiberInnen</font>: Bitte vorm
				Abschicken von Drohmails unsere
				<a href="%(proxyURL)s#wut">kl&auml;renden Worte</a> lesen.  Danke.</p>
				<p><a href="%(sourceURL)s">Unschwob</a></body>"""%{
					"proxyURL": self.proxyURL,
					"sourceURL": self.sourceURL,
				}
		else:
			return "</body>"

	def _raiseForbidError(self):
		"""(Hilfsfunktion, damit wir nicht jedes Mal wieder das gleiche
		Gewsch tippen mssen).
		"""
		raise common.PermissionError("Zielurl ist gesperrt -- "
				"Dies geschah vermutlich auf Wunsch des Betreibers")

	def _getDoc(self):
		"""holt sich das entfernte Dokument, bersetzt es und murkst ein
		bisschen mit den Headern rum.  Das sollte auch irgendwann mal
		schner werden.

		Als Seiteneffekt werden self.data, self.content_type und self.sourceURL
		gesetzt.
		"""
		form = self.request.form
		parlist = form.keys()
		try:
			parlist.remove(self.processor)
		except ValueError:
			pass
		if parlist:
			# Eventuelle Argumente hinbasteln
			try:
				formTag = "?"+"&".join(map(lambda a, form=form:
					a+"="+urllib.quote_plus(form[a].value), parlist))
			except AttributeError:
				raise common.Error("Irgendwas war mit der Parameterliste nicht"+\
					" in Ordnung -- ein Wert fehlte.  POST-Requests, das nur "+\
					"nebenbei, machen wir so oder so nicht.")
		else:
			formTag = ""

		if checkPermission(self.sourceURL):
			self._raiseForbidError()

		# Jetzt holen wir den Kram und sehen, dass wir bei Pannen so genau
		# wie mglich sagen, was schief gelaufen ist.
		try:
			u = urllib2.urlopen(workAroundUrllibBug(self.sourceURL+formTag))
		except urllib2.URLError, msg:
			self.request.log("%s, %s, %s"%(self.sourceURL, formTag, msg))
			try:
				raise common.Error("Mit %s ist etwas nicht in Ordnung: %s."%(
					self.sourceURL, str(msg.reason)))
			except AttributeError:
				raise common.Error("Mit %s ist etwas nicht in Ordnung: %s.")%(
					self.sourceURL, str(msg))
		except urllib2.HTTPError, msg:
			raise common.Error("%s hat ein Problem auf HTTP-Ebene: %s"%self.sourceURL,
				str(msg))
		except IOError, msg:
			if isinstance(msg.strerror, socket.error):
				self.request.log("urlopen sagte: %s, url=%s"%(str(msg),
					self.sourceURL+formTag))
				raise common.Error("Die Maschine am anderen Ende meldet sich nicht."
					" Luft da auch wirklich ein http-Daemon?")
			else:
				raise common.Error("Ich finde %s nicht"%self.sourceURL)
		except (ValueError, AssertionError):
			raise common.Error("Die URL %s ist "
				"keine URL -- oder wenn, verstehe ich sie nicht.  Der "
				"wahrscheinlichste Grund fr diesen Fehler ist Javascript, "
				"das ich nicht verstehe.  Vielleicht hilft es, Javascript "
				"abzuschalten."%self.sourceURL)
		except timeoutsocket.Timeout:
			self.request.log("Timeout...")
			raise common.Error("Die Maschine am anderen Ende hat sich"
				" zu lange nicht gemeldet.  Ich gebe auf.")
		except socket.error, msg:
			raise common.Error("Der Zugriff auf %s klappt aufgrund"
				" von tiefgrndigen technischen Problemen (namentlich: %s) nicht")%(
				self.sourceURL, str(msg))

		# Jetzt werfen wir noch einen schnellen Blick auf die Header.
		# TODO: Mehr von den Headern durchschleusen
		info = u.info()
		self.isData = info.getmaintype()!="text"
		if self.isData:
			raise Redirect(self.sourceURL)
		self.data = u.read(100000)
		self.sourceURL = u.geturl()
		self.content_type = "%s/%s"%(info.getmaintype(), info.getsubtype())
		u.close()
	
	def getData(self):
		return self.data
	
